/**
 * @file
 * Shadow manager class implementation.
 *
 * @author Tom "Shamot" Burian
 */
/* ---------------------------------------------------------------------------
 * Authors:      Tom Burian (tburian _AT centrum.cz)
 * Thanks to:    PCJohn (peciva _AT fit.vutbr.cz)
 * Contributors: 
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 * ---------------------------------------------------------------------------
 */

//----------------------------------------------------------------------------
// INCLUDEs
//----------------------------------------------------------------------------

#include "ShadowManager.h"
#include "SoShadowCaster.h"
#include "SoShadowVolume.h"


#ifdef _WIN32
  #include <windows.h> // needed by gl.h
#endif

// Must be linked with glu32.lib opengl32.lib.
#include <GL/gl.h>

//----------------------------------------------------------------------------
// IMPLEMENTATION
//----------------------------------------------------------------------------

float CShadowManager::envIntensity = 0.0f;

/**
 * 
 * Constructor.
 *
 * Needs to know the pointers to the user's scene graph and
 * a global environment node. 
 *
 * If environment node is NULL, Shadow Manager will try to look for
 * it under the scene root first and if none is fount, it
 * creates its own. The darkness of shadows can be controlled trough
 * the scene ambient light intensity of the user's SoEnvironment node.
 *
 * User MUST ENABLE STENCIL BUFFER to see shadows on his/her own by 
 * \code viewer->setStencilBuffer(TRUE); \endcode
 *
 * @param sr Pointer connection to the scene root node.
 * @param env Pointer connection to the scene global environment node.
 */
CShadowManager::CShadowManager(SoSeparator *sr, SoEnvironment * env)
    : sceneRoot(sr),
      environment(env)
{

  if ( this->sceneRoot == NULL )
  {
    this->printError("Your scene root node is NULL.");
  }
  else
  {
    fprintf(stdout,"\nShadows generated by Shadow Engine v %s\n",SE_VERSION);
    sceneRoot->ref(); // pokud se v konstruktoru nevola createShadowScene
                      // musi se ref  
  }

  first = TRUE;
  // auto increment indices
  objectIndex = 0;
  lightIndex = 0;

  // default SM settings
  settings = new struct TShadowSettings;
  settings->method = zpass;
  settings->dropShadows = TRUE;
  settings->fDepth = 150.f;  
  this->method2algorithm();
  ONDEBUG( this->printSettings() );

  // casters' root node
  shadowCastersRoot = new SoSeparator;
  shadowCastersRoot->setName("ShadowCastersRoot");
  shadowCastersRoot->ref();

  // whole shadow scene root
  shadowSceneRoot = new SoSeparator;
  shadowSceneRoot->setName("ShadowSceneRoot");
  shadowSceneRoot->ref();

  // volumes subgraph root 
  shadowLightsRoot = new SoSeparator;
  shadowLightsRoot->setName("shadowLightsRoot");
  shadowLightsRoot->ref();

  switchNode = new SoSwitch;
  sceneNoShadows = new SoSeparator;
  sceneWithShadows = new SoSeparator;
  sceneWithShadowVolumes = new SoSeparator;

  // the order of these children is important for the switchNode
  switchNode->addChild(sceneWithShadows);       // 0
  switchNode->addChild(sceneNoShadows);         // 1 
  switchNode->addChild(sceneWithShadowVolumes); // 2  

  // set default environment node
  if ( environment == NULL )
  {
    ONDEBUG( printf("look for  env node\n") );
    // try to find the node
    environment = this->findEnvironmentNode();
    if ( environment == NULL )
    {      
      ONDEBUG( printf("my own env\n") );
      environment = new SoEnvironment;      
      // default amb. light color and intensity (defaults as in Coin)
      environment->ambientColor.setValue(SbColor(1.f, 1.f, 1.f));

      environment->ambientIntensity.setValue(0.2f);
      this->shadowSceneRoot->addChild(environment);      
    }       
  }
  environment->ref();

  shadowSceneRoot->addChild(switchNode); // after env node
  this->setSceneType(CShadowManager::SHADOWS); // default
}


/**
 * Destructor.
 */
CShadowManager::~CShadowManager()
{
  ONDEBUG( printf("--CShadowManager::destructor()\n") );

  if ( this->sceneRoot != NULL )
    sceneRoot->unref();

  environment->unref();

  shadowCastersRoot->removeAllChildren();
  shadowCastersRoot->unref(); 
  shadowCastersRoot = NULL;

  shadowLightsRoot->removeAllChildren();
  shadowLightsRoot->unref(); 
  shadowLightsRoot = NULL;

  shadowSceneRoot->removeAllChildren();
  shadowSceneRoot->unref();
  shadowSceneRoot = NULL;

  switchNode = NULL;
}


/**
 * Recompute shadow volumes for all objects and all lights.
 */
void CShadowManager::createShadows()
{
  ONDEBUG( printf("--CShadowManager::createShadows()\n") );

  int cnt = this->shadowLightsRoot->getNumChildren();
  ONDEBUG( printf("--CShadowManager:: SVN cnt: %d\n",cnt) );
  for (int i = 0; i < cnt; i++) // create SV for one caster
  {
    SoShadowLight * slight = static_cast<SoShadowLight *>(this->shadowLightsRoot->getChild(i));
    slight->createShadowVolumes();
  }
  
  if ( this->sceneRoot != NULL )
    this->sceneRoot->touch();
}


/**
 * Sets which type of scene will be shown.
 *
 * @param type Enum of scene types.
 */
void CShadowManager::setSceneType(enum CShadowManager::SCENE_TYPE type)
{
  // the type is the order of node to traverse
  this->switchNode->whichChild.setValue(type);
}


/**
 * Set up whole shadow manager settings in one step.
 *
 * @param s New settings data structure.
 */
void CShadowManager::setManager(TShadowSettings * s)
{
  settings = s;  
  this->method2algorithm();
  this->createShadows();  
}


/**
 * Read shadow manager settings.
 *
 * @return Settings data structure.
 */
const struct TShadowSettings * CShadowManager::readSettings()
{
  return this->settings;
}


/**
 * Set shadow computation method.
 *
 * @param newMethod A new method.
 */
void CShadowManager::setMethod(enum TShadowMethod newMethod)
{  
  enum TShadowAlgorithm old = this->algorithm;
  this->settings->method = newMethod;
  this->method2algorithm();
  if ( old != this->algorithm ) 
  {
    ONDEBUG ( printf("setMethod: new algorithm => createShadows\n"); );
    this->createShadows();
  }
  else
  {
    ONDEBUG ( printf("setMethod: algorithm is not updated\n"); );
  }
  ONDEBUG ( printSettings() );
}


/**
 * Set shadow volumes depth.
 *
 * @param newfDepth A new SV depth.
 */
void CShadowManager::setSVDepth(float newfDepth)
{  
  this->settings->fDepth = newfDepth;
  this->createShadows();
  ONDEBUG ( printSettings() );
}


/**
 * Print shadow manager's settings.
 */
void CShadowManager::printSettings()
{ 
  char * a = "";
  char * m = "";
  char * ds = "";
  switch (settings->method) 
  {
    case zpass:   m = "z-pass";  break;
    case zfail:   m = "z-fail";  break;
    case autoset: m = "autoset"; break;
    default: m = "unknown"; break;
  }
  algorithm == zpassAlg ? a = "z-pass alg." : a = "z-fail alg.";
  settings->dropShadows ? ds = "shadows" : ds = "shadow volumes";  

  printf("\nSM Settings:\n");
  printf("  method:    %s\n", m);
  printf("  algorithm: %s\n", a);
  printf("  show:      %s\n", ds);
  printf("  far depth: %.2f\n\n", settings->fDepth);
}


/**
 * Get shadow volumes computing algorithm.
 *
 * @return Pointer to algorithm. Do not change the value, use method instead.
 */
TShadowAlgorithm * CShadowManager::getAlgorithm()
{
  return &algorithm;
}


/**
 * Set shadow algorithm according to the chosen method.
 */
void CShadowManager::method2algorithm()
{  
  ONDEBUG(printf("CShadowManager::method2algorithm\n"));
  switch ( settings->method )
  {
    case zpass: algorithm = zpassAlg; break;

    case zfail: algorithm = zfailAlg; break;

    case autoset: // TODO: determine algoritm on fly according camera position
                  // zfail is for now only
      algorithm = zfailAlg;
    break;

    default: algorithm = zfailAlg; break;
  }
}


/**
 * Disable objects' sensors.
 * 
 * This method is useful when an user want to change something (that is normally) 
 * detected by sensors and causes re-computation of shadows) in his existing
 * scene graph and do not want to re-compute shadows. Just disable sensors, 
 * do what you want and enable sensors afterwards.
 */
void CShadowManager::disableSensors()
{
  // for every caster create and add a new volume node
  int cnt = this->shadowCastersRoot->getNumChildren();
  ONDEBUG( printf("  disableSensors: SVN has %d children:\n",cnt));
  if ( cnt > 0 ) {  
    for (int i = 0; i < cnt; i++ ) {
      SoShadowCaster * caster = (SoShadowCaster *) this->shadowCastersRoot->getChild(i);
      caster->getMoveSensor()->detach();
      caster->getShapeSensor()->detach();
    } 
  } 
}


/**
 * Enable objects' sensors.
 */
void CShadowManager::enableSensors()
{
  // for every caster create and add a new volume node
  int cnt = this->shadowCastersRoot->getNumChildren();
  ONDEBUG( printf("  enableSensors: SVN has %d children:\n",cnt));
  if ( cnt > 0 ) {  
    for (int i = 0; i < cnt; i++ ) {
      SoShadowCaster * caster = (SoShadowCaster *) this->shadowCastersRoot->getChild(i);
      caster->getMoveSensor()->attach(caster->path);
      caster->getShapeSensor()->attach(caster->objectRoot);
    } 
  }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


/**
 * Get environment node registered in Shadow Manager.
 *
 * @return Pointer to the global environment node.
 */
SoEnvironment * CShadowManager::getEnvironmentNode()
{
  return this->environment;
}


/**
 * Sets global ambient light intensity.
 * Default value of Shadow Manager's own environment node is 0.2.
 *
 * @param val Float number between 0 and 1.
 */

void CShadowManager::setAmbientIntensity(float val)
{
  this->environment->ambientIntensity.setValue(val);
}

/**
 * Looks for  user's environment node in the user's scene.
 * 
 * If the node is found in the user's scene graph (first occurence only), 
 * its pointer is passed to the Shadow Manager. 
 * There should be only one instance of SoEnvironment 
 * in the whole user's scene graph. It is used to set the 
 * darkness of shadows according to the ambient light intensity.
 */
SoEnvironment * CShadowManager::findEnvironmentNode()
{
  if ( this->sceneRoot != NULL )
  {
    SoSearchAction * sa = new SoSearchAction;
    sa->setType(SoEnvironment::getClassTypeId());
    sa->setSearchingAll(FALSE);
    sa->apply(this->sceneRoot);

    SoPath * nodePath = sa->getPath();
    if ( nodePath == NULL ) {
      ONDEBUG( printf("SoEnvironment not FOUND\n") );
      return NULL;
    }
    else
    {
      if (nodePath->getTail()->isOfType(SoEnvironment::getClassTypeId()))
      {
        ONDEBUG( printf("SoEnvironment is FOUND !!!\n") );
        SoEnvironment * env = (SoEnvironment*) nodePath->getTail();
        return env;
      }
    }
  }
  return NULL;  
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

/**
 * Action called when any light changes its position.
 *
 * @param userdata Pointer to the shadow volume node.
 * @param s Light sensor.
 */
void CShadowManager::lightMoveCallback(void * userdata, SoSensor *s)
{
  ONDEBUG( printf("CShadowManager::lightMoveCallback for one light\n") );

  SoShadowLight * slight = static_cast<SoShadowLight *> (userdata);
  slight->createShadowVolumes();
}


/**
 * Adds a new light to the scene light manager.
 *
 * @param light One scene light.
 */
void CShadowManager::addLight(SoLight * light)
{  
  ONDEBUG( printf("addLight:\n") );
  int i = 0;
  int cnt = 0;

  // check, if the light is in the scene graph
  SoSearchAction * sa = new SoSearchAction;
  sa->setNode(light);
  sa->setSearchingAll(FALSE);
  sa->apply(this->sceneRoot);

  SoPath * nodePath = sa->getPath(); // tato cesta by se dala vyuzit i v senzoru napojenym na 
                                     // cestu
  if ( nodePath == NULL ) { // light is not in the scene graph, do not add it
    printError("addLight (warning): Light was not found in the scene graph.\nTherefore it was not added in the Shadow Manager\n");
  }
  else
  {
    bool found = FALSE;

    // for every caster create and add a new volume node
    // it has a sense to test if the light is not already in SM
    cnt = this->shadowLightsRoot->getNumChildren();
    if ( cnt > 0 )
    {  
      for (int i = 0; i < cnt; i++ )
      {
        SoShadowLight * slight = (SoShadowLight *) this->shadowLightsRoot->getChild(i);
        if ( slight->getLight() == light )
        {
          found = TRUE;
        }
      }
    }

    if ( !found ) // light is not there
    {

      // volume node name
      SbString lindex(lightIndex);
      SbString lname = "ShadowLight_";
      lname += lindex;
      ONDEBUG( printf("  sh light name: %s\n",lname.getString()) );

      light->ref(); // test -> presunout az do SoShL nebo nechat v SM?
      SoShadowLight * slight = new SoShadowLight(this, light);
      slight->setName(lname);
      slight->setLightName(lightIndex);

      // new sensor
      SoNodeSensor * lightSensor = new SoNodeSensor(this->lightMoveCallback, slight);
      lightSensor->attach((SoNode *) light);
      slight->setSensor(lightSensor);

      this->shadowLightsRoot->addChild(slight);

      // create shadow volumes for this light and all existing casters
      int cntc = this->shadowCastersRoot->getNumChildren();
      ONDEBUG( printf("  add light: num. casters: %d\n",cntc));
      if ( cntc > 0 ) {  
        for (int i = 0; i < cntc; i++ ) {
          SoShadowCaster * caster = (SoShadowCaster *) this->shadowCastersRoot->getChild(i);

          // 2 indexy
          // volume node name
          SbString cindex(caster->getCasterName());
          SbString vname = "SV_L";
          vname += lindex;
          vname += "_C";
          vname += cindex;
          ONDEBUG( printf("  sh light name: %s\n",vname.getString()) );

          //new volume node
          SoShadowVolume * volume = new SoShadowVolume(slight, caster->getCasterName());
          volume->setName(vname);
          slight->addVolume(volume);
          
          ONDEBUG( fprintf(stdout,"  child %d: %s\n",i,caster->getName().getString()) );
        } 
      } 

      lightIndex++;
      // force shadows re-computation
      this->createShadows();
    } else {
      this->printError("Light is already present.");
    }
  }
  ONDEBUG(
    printf(" num. lights %d \n", this->shadowLightsRoot->getNumChildren() );
  );
  this->checkLights();  
  ONDEBUG( printf("addLight end\n\n") );  
}


/**
 * Removes light from Shadow Manager and deletes all shadows that
 * are cast by this light.
 * 
 * Note: Remove light from Shadow Manager using this method and 
 * after that you can remove it from your scene graph. 
 * Otherwise it will not work properly.
 *
 * @param light One scene light.
 */
void CShadowManager::removeLight(SoLight * light)
{ 
  ONDEBUG( printf("removeLight:\n") );

  int i = 0;

  if ( light == NULL ) 
  {
    this->printError("Light is NULL.");       
  }
  else
  {  
    int cnt = this->shadowLightsRoot->getNumChildren();
    if ( cnt > 0 )
    {  
      for ( i = 0; i < cnt; i++ )
      {
        SoShadowLight * slight = (SoShadowLight *) this->shadowLightsRoot->getChild(i);
        if ( slight->getLight() == light )
        {          
          this->shadowLightsRoot->removeChild(i);
          // force shadows re-computation
          this->createShadows();          
          light->unref();  
          break;
        }
      }
      if ( i == cnt ) 
        this->printError("Light is not in SM.");   
    }
  }

  ONDEBUG(
    checkLights();
    printf("  lights %d \n", this->shadowLightsRoot->getNumChildren() );
    printf("removeLight end\n\n");
  );
}


/**
 * Check all lights.
 */
void CShadowManager::checkLights()
{  
  bool found = FALSE;

  int cnt = this->shadowLightsRoot->getNumChildren();
  if ( cnt > 0 )
  {  
    for (int i = 0; i < cnt; i++ )
    {
      SoShadowLight * slight = (SoShadowLight *) this->shadowLightsRoot->getChild(i);

      slight->getLight()->on.enableNotify(FALSE);
      SbBool s = slight->getLight()->on.getValue(); // ligtht's status
      char * status = "";
      s ? status = "on" : status = "off";

      char * typ = "";
      if (slight->getLight()->getTypeId() == SoDirectionalLight::getClassTypeId() )
        typ = "SoDirectionalLight";
      else if (slight->getLight()->getTypeId() == SoDirectionalLightManip::getClassTypeId() )
        typ = "SoDirectionalLightManip";
      else if (slight->getLight()->getTypeId() == SoPointLight::getClassTypeId() )
        typ = "SoPointLight";
      else if (slight->getLight()->getTypeId() == SoPointLightManip::getClassTypeId() )
        typ = "SoPointLightManip";
      else if (slight->getLight()->getTypeId() == SoSpotLight::getClassTypeId() )
        typ = "SoSpotLight";
      else if (slight->getLight()->getTypeId() == SoSpotLightManip::getClassTypeId() )
        typ = "SoSpotLightManip";
      else 
        typ = "Unknown light type";

      ONDEBUG( fprintf(stdout,"  light %d: %s %s\n",i,status,typ) );
    }
  }

}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

/**
 * Action called when something has changed in the path to the object node.
 * For example, the light position has changed or any node has 
 * been added to the scene graph before object node.
 *
 * The shadow volume of the object must be recalculated.
 *
 * @param userdata Shadow caster.
 * @param s Object sensor.
 */
void CShadowManager::objectMoveCallback(void * userdata, SoSensor *s)
{
  ONDEBUG( printf("@@@caster path CB\n") );

/*
  o co tady jde: 
  1) pri zmene geometrie objektu (polygony) vytvorit znovu face model objektu
     sem spada i zmena scale -> v tom pripade by se model nemusel prepocitavat cely,
     jen k nemu pridat scale

     potom prepocitat i stinove teleso

  2) zmeni-li se jen poloha/natoceni objektu ve scene, jen prepocitat stinove teleso
    
  Problem: jak tyto dve veci od sebe rozlisit?
  
*/

  SoShadowCaster * caster = static_cast<SoShadowCaster *>(userdata);  

  ONDEBUG( printf("@@@caster path CB called by %s\n", caster->getName().getString()) );

  // we know the sensor is really a data sensor
  SoDataSensor *mySensor = (SoDataSensor *)s;
    
  SoNode *changedNode = mySensor->getTriggerNode();
  SoField *changedField = mySensor->getTriggerField();
  
  if ( changedNode != NULL ) {
    ONDEBUG( printf("  The node named '%s' changed\n", changedNode->getName().getString()) );
    if ( changedNode->getTypeId().isDerivedFrom(SoTransformation::getClassTypeId()) )
    {
      // translation or rotation has changed
      if ( !changedNode->isOfType(SoScale::getClassTypeId()) ) {
        ONDEBUG( printf("  create shadow volume for the caster (not a scale tr.)\n") );
        caster->setMoved(TRUE);
        caster->createShadowVolumes();        
      } else { // scale has changed
        ONDEBUG( printf("  create shadow volume: scale transform\n") );
        caster->createFaceModel();
        caster->createShadowVolumes();
      }
    }
    else
    {
      // something in the scene graph has changed (some node was removed or added)
      // or touch called
      // re-compute face model (for sure) and shadow volumes
      // this is not affected by moving with the light
      if ( changedField == NULL )
      {
/*
        ONDEBUG( printf("  changedNode is NOT NULL but changedField is NULL, no SoTransformation \n") );
        caster->createFaceModel();
        caster->createShadowVolumes();
*/
      }
/*
// toto je strasne divny, reakce na animaci
      else
      {
        if ( !changedNode->isOfType(SoLight::getClassTypeId()) ) {
          ONDEBUG( printf("  changedNode is NOT NULL and changedField is NOT NULL as well, no SoTransformation, no SoLight \n") );
//        caster->createFaceModel();
          caster->createShadowVolumes();
        }
      }
*/
    }
  } 
  else 
  {
    ONDEBUG( printf("  changedNode is NULL \n") );
  }

  ONDEBUG( printf("@@@caster path CB done\n") );
}


/**
 * Action called when the object changes its shape.
 *
 * @param userdata Shadow caster.
 * @param s Object node sensor.
 */
void CShadowManager::objectShapeCallback(void * userdata, SoSensor *s)
{
  ONDEBUG( printf("###caster shape CB\n") );

  SoShadowCaster * caster = static_cast<SoShadowCaster *>(userdata);

  // we know the sensor is really a data sensor
  SoDataSensor * mySensor = (SoDataSensor *) s;
    
  SoNode *changedNode = mySensor->getTriggerNode();
  SoField *changedField = mySensor->getTriggerField();
    
  if ( changedNode != NULL ) {
    ONDEBUG( printf("  The node named '%s' changed\n", changedNode->getName().getString()) );

    if (changedField == NULL) // can be caused by add child or touch
    {                         // it is called when a new object is added

      ONDEBUG( printf("  changedField == NULL\n") );
      if ( changedNode->isOfType(SoSeparator::getClassTypeId()) ) {
        ONDEBUG( printf("  added %d \n",caster->getCasterName()) );
        // recompute caster model
        caster->createFaceModel();
        // prepocitat SV
        caster->createShadowVolumes();
      }
    }
  }
  else
  {
    ONDEBUG( printf("  changedNode is NULL in shape CB\n") );
  }
  ONDEBUG( printf("###caster shape CB\n") );
}

/**
 * Looks for a SoShadowCaster node for this object inside shadow manager.
 *
 * @param object A scene object node.
 *
 * @return -1 if object was not found, object name index otherwise
 */
int CShadowManager::findObject(SoSeparator * object)
{
  // check, if the object is not already in the shadow manager
  int found = -1;
  int cnt = this->shadowCastersRoot->getNumChildren();
  if ( cnt > 0 )
  {
    for (int i = 0; i < cnt; i++ )
    {
      SoShadowCaster * caster = (SoShadowCaster *) this->shadowCastersRoot->getChild(i);
      if ( caster->getObjectNode() == object )
      {
        ONDEBUG( printf("findObject: Object is already in SM: %d\n", caster->getCasterName()) );
        return caster->getCasterName();
      }    
    } 
  }
  ONDEBUG( printf("findObject: Object is not in SM\n") );
  return found;  
}


/**
 * Adds a new object to the scene object manager.
 *
 * @param object A scene object node. 
 *
 * An object is assumed to be a SoSeparator node under which 
 * a complete object definition (coordinates, indices, materials...)
 * is placed. Any change below this node (a shape change) is detected
 * and shadows for this object are updated. Object's transformations 
 * placed in the scene graph above the object root node are detected as well.
 *
 * Each user object can be added into the Shadow Manger only once!
 */
void CShadowManager::addObject(SoSeparator * object)
{   
  ONDEBUG( printf("add Object\n") );
  int i = 0;

  // check, if the object is in the scene graph  
  SoSearchAction * sa = new SoSearchAction;
  sa->setNode(object);
  sa->setSearchingAll(FALSE);
  sa->apply(this->sceneRoot);

  // a path from scene root to the object node to detect any transformation changes
  SoPath * nodePath = sa->getPath();
  if ( nodePath == NULL ) { // object is not in the scene graph, do not add it
    printError("addObject (warning): Object was not found in the scene graph.\nTherefore it was not added in the Shadow Manager.\n");
  }
  else // object is in the scene graph
  { 
    if ( findObject(object) < 0 ) // add object
    { 
      // add new casting object node
      SoShadowCaster * caster = new SoShadowCaster(this, object);
      caster->setCasterName(objectIndex);
      this->shadowCastersRoot->addChild(caster);
      objectIndex++;
  
      // ref the object, for sure
      object->ref();
                  
      // create a new object sensor to detect location changes
      SoPathSensor * objectSensor = new SoPathSensor(this->objectMoveCallback, caster);
      objectSensor->setPriority(0); // important
      objectSensor->attach(nodePath);
      caster->setMoveSensor(objectSensor);

      caster->path = nodePath;

      // create a new object sensor to detect shape changes
      SoNodeSensor * objectSensorC = new SoNodeSensor(this->objectShapeCallback, caster);
      objectSensorC->attach((SoNode *) object);
      objectSensorC->setPriority(0); // important
      caster->setShapeSensor(objectSensorC);

      // tohle by se melo dat do nejake metody
      // create volume nodes for all lights in SM and this object
      int cnt = this->shadowLightsRoot->getNumChildren();
      if ( cnt > 0 )
      {
        for (int i = 0; i < cnt; i++ )
        {
          SoShadowLight * slight = (SoShadowLight *) this->shadowLightsRoot->getChild(i);
      
          // 2 indexy
          // volume node name
          SbString cindex(caster->getCasterName());
          SbString lindex(slight->getLightName());
          SbString vname = "SV_L";
          vname += lindex;
          vname += "_C";
          vname += cindex;
          ONDEBUG( printf("  sh volume name: %s\n",vname.getString()) );

          //new volume node
          SoShadowVolume * volume = new SoShadowVolume(slight, caster->getCasterName());
          volume->setName(vname);
          
          slight->addVolume(volume);
        }
      }

      ONDEBUG( printf("  child %d: %s\n",i,caster->getName().getString()) );
      ONDEBUG( printf("  add obj: volume added\n") );
      ONDEBUG( printf("  The new caster has %d children.\n", caster->getNumChildren()) );

      // very importatnt thing. this nottyfies the attached sensor and 
      // calls appropriate CB method which creates object's model and shadows
      // also causes redraw of the scene, i.e. shadow appears in it
      object->touch();
      ONDEBUG( this->printMsg("  Object was added.") );

    } else {
      this->printError("Object is already present.");
    }    
  }
  ONDEBUG( checkObjects() );
  ONDEBUG( printf("add object done\n\n") );
}


/**
 * Removes the first occurence of the object from Shadow Manager.
 * 
 * When object is found in the Shadow Manager, its shadows are
 * destroyed. Also the sensor is detached and deleted.
 *
 * @param object Pointer to a scene object node.
 */
void CShadowManager::removeObject(SoSeparator * object)
{ 

  ONDEBUG( printf("remove object\n") );
  // check, if the object is not already in the shadow manager
  int found = -1;
  int cnt = this->shadowCastersRoot->getNumChildren();
  if ( cnt > 0 ) 
  {  
    for (int i = 0; i < cnt; i++ )
    {
      SoShadowCaster * caster = (SoShadowCaster *) this->shadowCastersRoot->getChild(i);
      if ( caster->getObjectNode() == object )
      {
        ONDEBUG( printf("  removing caster: %d\n", caster->getCasterName()) );
        found = caster->getCasterName();

        // delete shadow volumes of this caster and each light
        int cntl = this->shadowLightsRoot->getNumChildren();
        for (int j = 0; j < cntl; j++ )
        {
          SoShadowLight * slight = (SoShadowLight *) this->shadowLightsRoot->getChild(j);
          slight->deleteVolume(caster->getCasterName());
        }

        this->shadowCastersRoot->removeChild(i);        
        // unref the object, was ref()ed
        object->unref();

        break; // to prevent failure; remove the first occurence only
        // musi tady byt break, ptz pocet deti se snizi o 1 a po dstraneni 1 se
        // saha za konec () cnt se neupdatuje, musel by se pouzit while nejak
      }    
    } 
  }  
  if ( found < 0 ) { 
    this->printError("removeObject: object not found.");
  }
  ONDEBUG( checkObjects() );
  ONDEBUG( printf("remove object done\n\n") );

}


/**
 * Forces recomputation of shadows of an object.
 * This method is useful when something changes in the scene
 * graph that affects object's shadows and none of
 * Shadow Manager's internal callbacks is trigerred.
 *
 * @param object Pointer to a scene object node.
 */
void CShadowManager::createObjectsShadows(SoSeparator * object)
{ 
  ONDEBUG( printf("recreateObjectsShadows\n") );
  // check, if the object is not already in the shadow manager
  int found = -1;
  int cnt = this->shadowCastersRoot->getNumChildren();
  if ( cnt > 0 ) 
  {  
    for (int i = 0; i < cnt; i++ )
    {
      SoShadowCaster * caster = (SoShadowCaster *) this->shadowCastersRoot->getChild(i);
      if ( caster->getObjectNode() == object )
      {
        found = 1;
        caster->createFaceModel();
        caster->createShadowVolumes();
      }    
    } 
  }  
  if ( found < 0 ) { 
    this->printError("recreateObjectsShadows: object not found.");
  }
  ONDEBUG( printf("recreateObjectsShadows done\n\n") );
}


/**
 * Check all objects under SVN node.
 */
void CShadowManager::checkObjects()
{  
  int cnt = this->shadowCastersRoot->getNumChildren();
  ONDEBUG( fprintf(stdout,"\nSVN has %d children:\n",cnt) );
  if ( cnt > 0 ) {  
    for (int i = 0; i < cnt; i++ ) {
      SoShadowCaster * caster = (SoShadowCaster *) this->shadowCastersRoot->getChild(i);
      ONDEBUG( fprintf(stdout,"  child %d: %s\n",i,caster->getName().getString()) );
    } 
  } 
  ONDEBUG( fprintf(stdout,"\n") );
}


/**
 * Re-computes shadow volumes of a selected caster.
 *
 * @param index Caster name.
 */
void CShadowManager::createCasterVolumes(int index)
{
  int cnt = this->shadowLightsRoot->getNumChildren();
  if ( cnt > 0 )
  {
    for (int i = 0; i < cnt; i++ )
    {
      SoShadowLight * slight = (SoShadowLight *) this->shadowLightsRoot->getChild(i);

      int vcnt = slight->getNumVolumes();
      if ( vcnt > 0 )
      {
        for (int j = 0; j < vcnt; j++ )
        {      
          //new volume node
          SoShadowVolume * volume = (SoShadowVolume *) slight->getVolume(j);
          if ( volume->getCasterName() == index )
          {          
            volume->create();
          }
        }
      }
    }
  }
}


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

/**
 * Creates scene graph with shadows.
 * 
 * Creates scene graph with shadows under the CShadowManager.
 */
void CShadowManager::createShadowScene()
{  
  ONDEBUG( printf("creating shadow scene graph...\n") );

// rendering callbacks
  // scene without shadows
  SoCallback * nslightsOffCallback = new SoCallback;
  nslightsOffCallback->setCallback(nslightsOffCB, this->shadowLightsRoot);

  SoCallback * nslightsOnCallback = new SoCallback;
  nslightsOnCallback->setCallback(nslightsOnCB, this->shadowLightsRoot);

  SoCallback * nsambientOffCallback = new SoCallback;
  nsambientOffCallback->setCallback(nsambientOffCB, this->environment);

  SoCallback * nsambientOnCallback = new SoCallback;
  nsambientOnCallback->setCallback(nsambientOnCB, this->environment);

  // scene with shadows
  SoCallback * initCallback = new SoCallback;
  initCallback->setCallback(initCB, this->shadowLightsRoot);
  
  SoCallback * endCallback = new SoCallback;
  endCallback->setCallback(endCB, this->shadowLightsRoot);

  ONDEBUG( this->printSettings() ); 

  if ( this->sceneRoot != NULL )
  {

    // what to show

    // only user scene
    this->sceneNoShadows->addChild(nslightsOffCallback);    
    this->sceneNoShadows->addChild(this->sceneRoot);
    this->sceneNoShadows->addChild(nslightsOnCallback);  
    this->sceneNoShadows->addChild(nsambientOffCallback);  
    this->sceneNoShadows->addChild(this->sceneRoot);
    this->sceneNoShadows->addChild(nsambientOnCallback);

    // scene with shadows
    
    // create top part of shadow scene graph
    SoSeparator * aboveShadowVloumesRoot = new SoSeparator;
    this->sceneWithShadows->addChild(aboveShadowVloumesRoot);

    aboveShadowVloumesRoot->addChild(initCallback);
    aboveShadowVloumesRoot->addChild(this->sceneRoot);

    aboveShadowVloumesRoot->addChild(shadowLightsRoot);
    aboveShadowVloumesRoot->addChild(endCallback);

    // scene with shadow volumes ??? 

/*
    // create top part of shadow scene graph
    SoSeparator * aboveShadowVloumesRoot = new SoSeparator;
    shadowSceneRoot->addChild(aboveShadowVloumesRoot);

    aboveShadowVloumesRoot->addChild(initCallback);
    aboveShadowVloumesRoot->addChild(this->sceneRoot);

    aboveShadowVloumesRoot->addChild(shadowLightsRoot);
    aboveShadowVloumesRoot->addChild(endCallback);
*/
  }
}


/**
 * Returns the root node of the user scene with shadows.
 *
 * @return SoSeparator * Pointer to the root node of shadow scene. 
 */
SoSeparator * CShadowManager::getShadowSceneRoot()
{
  if ( first ) {
    this->createShadowScene();
    first = FALSE;
  }
  return this->shadowSceneRoot;
}


/**
 * Returns the root node of the user's scene.
 *
 * @return SoSeparator * Pointer to the root node of scene. 
 */
SoSeparator * CShadowManager::getSceneRoot()
{
  return this->sceneRoot;
}


//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

/**
 * Prints information to stdout.
 *
 * @param msg Message.
 */
void CShadowManager::printMsg(const char *msg)
{
  fprintf(stdout,"SM: %s\n",msg);
}


/**
 * Error printing method.
 *
 * @param msg Error message.
 */
void CShadowManager::printError(const char *msg)
{
  fprintf(stderr,"SM Error: %s\n",msg);
}


/**
 * Export shadow scene graph into a file.
 *
 * @param filename Filename of exported scene. 
 */
void CShadowManager::export(const char * filename)
{
  SoOutput output;
  if ( output.openFile(filename) ) {
    SoWriteAction * writer = new SoWriteAction(&output);
    writer->apply(this->shadowLightsRoot);
    delete writer;
  }
  else
  {
    char * buf = NULL;
    sprintf(buf,"Can not open file '%s' for export.",filename);
    this->printError(buf);
  }
}


/*
//TODO: Load a shadow scene graph from a file.
void CShadowManager::import(const char * filename)
{
  if ( filename != NULL ) 
  {
    SoInput in;
    if (!in.openFile(filename)) {
      char * buf = NULL;
      sprintf(buf,"Can not open file '%s' for import.",filename);
      this->printError(buf);
    }

    SoSeparator * loadNode = SoDB::readAll(&in);
    loadNode = SoDB::readAll(&in);
    loadNode->setName("loadNode");
    if (loadNode == NULL) {
      char * buf = NULL;
      sprintf(buf,"Can not read file '%s' for import.",filename);
      this->printError(buf);
    }
    else
    {
      this->shadowCastersRoot->removeAllChildren();
      this->shadowLightsRoot->removeAllChildren();
      this->shadowLightsRoot->addChild(loadNode);
    }    
  }
}
*/











//////////////////////////////////////////////////////////////////////////////
// NO Shadow rendering callbacks
//////////////////////////////////////////////////////////////////////////////



/**
 * Rendering ambient CB.
 * 
 * Turns off lights.
 * 
 * @param userdata Pointer to the root of shadow lights.
 * @param action Pointer to the callback action.
 */
void CShadowManager::nslightsOffCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {

    ONDEBUG(printf("nslightsOffCB\n"));

    // depth
    glClearDepth(1.0f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);    

    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // ma vliv na barvu a text. souradnice
//    glEnable(GL_CULL_FACE);    
    
    // Clear all buffers
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glPushAttrib(GL_ENABLE_BIT); 

    glEnable(GL_LIGHTING);

    // turn off the lights and save each light status
    SoSeparator * lightRoot = (SoSeparator * ) userdata;
    int cnt = lightRoot->getNumChildren();
    for ( int i = 0; i < cnt; i++ )
    {
      SoShadowLight * slight = static_cast<SoShadowLight *>(lightRoot->getChild(i));
      slight->getLight()->on.enableNotify(FALSE);
      slight->status = slight->getLight()->on.getValue();
      slight->getLight()->on = FALSE;
      slight->getLight()->on.enableNotify(TRUE);
    }
   
  }
} 


/**
 * Rendering ambient end CB.
 * 
 * Turns on lights.
 * 
 * @param userdata Pointer to the root of shadow lights.
 * @param action Pointer to the callback action.
 */
void CShadowManager::nslightsOnCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {

    ONDEBUG(printf("nslightsOnCB\n"));

    // turn on the lights to its saved status
    SoSeparator * lightRoot = (SoSeparator * ) userdata;
    int cnt = lightRoot->getNumChildren();
    for ( int i = 0; i < cnt; i++ )
    {
      SoShadowLight * slight = static_cast<SoShadowLight *>(lightRoot->getChild(i));
      slight->getLight()->on.enableNotify(FALSE);
      slight->getLight()->on = slight->status;
      slight->getLight()->on.enableNotify(TRUE);
    }   
  }
} 


/**
 * Rendering no shadows light CB.
 *
 * Turns off and save ambient intensity value. Sets additive blending.
 *
 * @param userdata Pointer to environment node in the Shadow Manager.
 * @param action Pointer to the callback action.
 */
void CShadowManager::nsambientOffCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("nsambientOffCB\n"));    

    glBlendFunc(GL_ONE, GL_ONE);
    glEnable(GL_BLEND);    

    SoEnvironment * env = (SoEnvironment * ) userdata;

    // disable ambient light (additive blending)
    // (save ambient int. value)
    if ( env != NULL )
    {
      env->ambientIntensity.enableNotify(FALSE);
      envIntensity = env->ambientIntensity.getValue();
      env->ambientIntensity.setValue(0.0f);
      env->ambientIntensity.enableNotify(TRUE);
    }
  }
} 


/**
 * Rendering no shadow end CB.
 *
 * Closes the whole rendering action. Sets default value
 * of ambient light intensity.
 *
 * @param userdata Pointer to environment node in the Shadow Manager.
 * @param action Pointer to the callback action.
 */
void CShadowManager::nsambientOnCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("nsambientOnCB\n"));

    SoEnvironment * env = (SoEnvironment * ) userdata;

    // enable ambient light 
    if ( env != NULL )
    {
      env->ambientIntensity.enableNotify(FALSE);
      env->ambientIntensity.setValue(envIntensity);
      env->ambientIntensity.enableNotify(TRUE);
    }
    
    glDisable(GL_BLEND);
    glDepthFunc(GL_LEQUAL);
    glPopAttrib();
  }
} 


//////////////////////////////////////////////////////////////////////////////
// Shadow rendering callbacks
//////////////////////////////////////////////////////////////////////////////


/**
 * Rendering preset CB.
 *
 * Starts the whole rendering action. Saves light status values and 
 * cleares buffers.
 *
 * @param userdata Pointer to root of all lights in the Shadow Manager.
 * @param action Pointer to the callback action.
 */
void CShadowManager::initCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("initCB\n"));    

    // depth
    glClearDepth(1.0f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);    

    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST); // ma vliv na barvu a text. souradnice
    glEnable(GL_CULL_FACE);    
    
    // Clear all buffers
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glPushAttrib(GL_ENABLE_BIT); 

    glEnable(GL_LIGHTING);

    // turn off the lights and save each light status
    SoSeparator * lightRoot = (SoSeparator * ) userdata;
    int cnt = lightRoot->getNumChildren();
    for ( int i = 0; i < cnt; i++ )
    {
      SoShadowLight * slight = static_cast<SoShadowLight *>(lightRoot->getChild(i));
      slight->getLight()->on.enableNotify(FALSE);
      slight->status = slight->getLight()->on.getValue();
      slight->getLight()->on = FALSE;
      slight->getLight()->on.enableNotify(TRUE);
    }
  }
} 


/**
 * Rendering end CB.
 *
 * Closes the whole rendering action. Sets default values and 
 * resets the light state according to the saved values.
 *
 * @param userdata Pointer to root of all lights in the Shadow Manager.
 * @param action Pointer to the callback action.
 */
void CShadowManager::endCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("endCB\n"));
    
    // set back each light status 
    SoSeparator * lightRoot = (SoSeparator * ) userdata;
    int cnt = lightRoot->getNumChildren();
    for ( int i = 0; i < cnt; i++ )
    {
      SoShadowLight * slight = static_cast<SoShadowLight *>(lightRoot->getChild(i));
      slight->getLight()->on.enableNotify(FALSE);      
      slight->getLight()->on = slight->status;
      slight->getLight()->on.enableNotify(TRUE);
    }
    
    glDepthFunc(GL_LEQUAL);
    glPopAttrib();
  }
} 
