/**
 * @file
 * Shadow manager class definition.
 *
 * @author Tom "Shamot" Burian
 */
/* ---------------------------------------------------------------------------
 * Authors:      Tom Burian (tburian _AT centrum.cz)            
 * Thanks to:    PCJohn (peciva _AT fit.vutbr.cz)
 * Contributors: 
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 * ---------------------------------------------------------------------------
 */

#ifndef SHADOW_MANAGER__H
#define SHADOW_MANAGER__H

/** Debug information toggle. Print debug info if set to 1. */
#ifndef DODEBUG 
#define DODEBUG 0
#endif 
#if DODEBUG
/** Debug function that is mainly used for printing messages. */
#define ONDEBUG(x) x 
#else
/** Does nothing. */
#define ONDEBUG(x) 
#endif 

/** Show face computed model in the scene. */
#define SHOWFACEMODEL 0 

#ifndef SE_VERSION  /** Shadow Engine version. */
#define SE_VERSION "0.6.2"
#endif 

//----------------------------------------------------------------------------
// INCLUDEs
//----------------------------------------------------------------------------

#include "DataStructures.h"
//
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoCallback.h>
#include <Inventor/nodes/SoPolygonOffset.h>
#include <Inventor/nodes/SoTransformation.h>
#include <Inventor/nodes/SoEnvironment.h>
//
#include <Inventor/nodes/SoLight.h>
#include <Inventor/nodes/SoDirectionalLight.h> 
#include <Inventor/nodes/SoPointLight.h>
#include <Inventor/nodes/SoSpotLight.h> 
//
#include <Inventor/manips/SoDirectionalLightManip.h> 
#include <Inventor/manips/SoPointLightManip.h> 
#include <Inventor/manips/SoSpotLightManip.h> 
//
#include <Inventor/sensors/SoFieldSensor.h> 
#include <Inventor/sensors/SoNodeSensor.h> 
#include <Inventor/sensors/SoPathSensor.h>
//
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoGLRenderAction.h>
//
#include <Inventor/lists/SbList.h> 

//----------------------------------------------------------------------------
// IMPLEMENTATION
//----------------------------------------------------------------------------

/**
 * The Manager of shadows in Open Inventor scene.
 *
 * Provides simple API for adding shadows to any scene object.
 * Shadow Manager has methods to add lights and scene objects
 * for which shadows should be created. Automatically handles
 * lights or objects movements and shape changes of shadow 
 * casting objects.
 *
 * Compilation note: Must be linked with glu32.lib opengl32.lib.
 */
class CShadowManager
{

  public:
  
    /** Enumeration of showed scene types. */
    enum SCENE_TYPE {  
             SHADOWS = 0, /**< User scene with shadows. */  
               SCENE = 1  /**< User scene. */               
//      SHADOW_VOLUMES = 2  /**< User scene with shadow volumes. */ // to do
    };


  private:
    /** Shadow casters root node. */
    SoSeparator * shadowCastersRoot;
    /** Shadow scene root node. */
    SoSeparator * shadowSceneRoot;
    /** Root node of all scene shadow volumes. */
    SoSeparator * shadowLightsRoot;

    /** Scene root node. */
    SoSeparator * sceneRoot;
    /** Shadow settings data structure. */
    struct TShadowSettings * settings;
    /** Shadow algorithm. */
    enum TShadowAlgorithm algorithm;
    /** Auto incremental index (name) of the light. */
    int lightIndex;
    /** Auto incremental index (name) of the object. */
    int objectIndex;
    /** Is it the first call of create shadow scene?. */
    bool first;
    /** Environment node to set ambient light. */
    SoEnvironment * environment;
    /** Ambient light intensity saved value. */
    static float envIntensity;    

    /** Node to switch between scene types. */
    SoSwitch * switchNode;
    /** Root node of scene without shadows. */
    SoSeparator * sceneNoShadows;
    /** Root node of scene with shadows. */
    SoSeparator * sceneWithShadows;
    /** Root node of scene with shadow volumes. */
    SoSeparator * sceneWithShadowVolumes;

// shadows    
    void createShadowScene();        

// environment
    SoEnvironment * findEnvironmentNode();

// lights
    static void lightMoveCallback (void * userdata, SoSensor *);    
    void checkLights();

// shadow casters / objects
    static void objectMoveCallback(void * userdata, SoSensor * s);
    static void objectShapeCallback(void * userdata, SoSensor * s);
    int findObject(SoSeparator * object);
    void checkObjects();    
    
// rendering callbacks
    // without shadows    
    static void nslightsOffCB(void * userdata, SoAction * action);
    static void nslightsOnCB(void * userdata, SoAction * action);
    static void nsambientOffCB(void * userdata, SoAction * action);
    static void nsambientOnCB(void * userdata, SoAction * action);
    // shadows
    static void initCB(void * userdata, SoAction * action);
    static void endCB(void * userdata, SoAction * action);

// messages
    void printMsg(const char * msg);
    void printError(const char * msg);

  public:
    CShadowManager(SoSeparator * sr, SoEnvironment * env);
    ~CShadowManager();

// scene types
    void setSceneType(enum CShadowManager::SCENE_TYPE type);

// method
    enum TShadowAlgorithm * getAlgorithm();
    void method2algorithm();
    
// settings
    void setManager(TShadowSettings * s);
    const struct TShadowSettings * readSettings();
    void setMethod(enum TShadowMethod newMethod);
    void setSVDepth(float newfDepth);
    void printSettings();
    void enableSensors();
    void disableSensors();

// shadows
    void createShadows();
    void createCasterVolumes(int index);
    SoSeparator * getSceneRoot();
    SoSeparator * getShadowSceneRoot();

// environment
    SoEnvironment * getEnvironmentNode(); 
    void setAmbientIntensity(float val);

// scene lights
    void addLight(SoLight * light);
    void removeLight(SoLight * light);
    
// scene objects    
    void addObject(SoSeparator * object);
    void removeObject(SoSeparator * object);
    void createObjectsShadows(SoSeparator * object);

// export / import
    void export(const char * filename);
//    void import(const char * filename);
};


#endif
