/**
 * @file
 * Shadow casting object class implementation.
 *
 * @author Tom "Shamot" Burian
 */
/* ---------------------------------------------------------------------------
 * Authors:      Tom Burian (tburian _AT centrum.cz)
 * Thanks to:    PCJohn (peciva _AT fit.vutbr.cz)
 * Contributors:
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 * ---------------------------------------------------------------------------
 */

//----------------------------------------------------------------------------
// INCLUDEs
//----------------------------------------------------------------------------

#include "SoShadowCaster.h"

//----------------------------------------------------------------------------
// IMPLEMENTATION
//----------------------------------------------------------------------------

/**
 * Constructor.
 *
 * @param sm Pointer connection to the Shadow Manager.
 * @param or Pointer connection to the scene object node.
 */
SoShadowCaster::SoShadowCaster(CShadowManager * sm, SoSeparator * or)
    : manager(sm),
      objectRoot(or)
{  
  path = NULL;
  moveSensor = NULL;
  shapeSensor = NULL;
  nameIndex = -1;
  objectModel = new CObjectModel();  
  this->getTransformationMatrix();
  hasMoved = FALSE;
}


/**
 * Destructor.
 */
SoShadowCaster::~SoShadowCaster()
{
  // sensors must be detached !!!
  this->moveSensor->detach();
  this->shapeSensor->detach();
//  this->removeAllChildren(); // has no children
}


/**
 * Returns object node pointer.
 */
SoSeparator * SoShadowCaster::getObjectNode()
{
  return this->objectRoot;
}


/**
 * Sets caster name index and the name of the node.
 *
 * @param objectIndex Auto increment index name.
 */
void SoShadowCaster::setCasterName(int objectIndex)
{
  SbString index(objectIndex);
  SbString name = "Caster_";
  name += index;
  ONDEBUG( printf("-- caster name: %s\n",name.getString()) );
  this->nameIndex = objectIndex;
  this->setName(name);
}


/**
 * Gets caster name index.
 *
 * @return int Index (name) of the object.
 */
int SoShadowCaster::getCasterName()
{
  return this->nameIndex;
}


/**
 * Sets object move sensor.
 *
 * @param sensor Pointer to the PathSensor to the object node.
 */
void SoShadowCaster::setMoveSensor(SoPathSensor * sensor)
{
  this->moveSensor = sensor;
}


/**
 * Gets object move sensor.
 *
 * @return SoPathSensor * pointer to the object move sensor.
 */
SoPathSensor * SoShadowCaster::getMoveSensor()
{
  return this->moveSensor;
}


/**
 * Sets object moveSensor geom. change.
 *
 * @param sensor Pointer to the NodeSensor to the object node.
 */
void SoShadowCaster::setShapeSensor(SoNodeSensor * sensor)
{
  this->shapeSensor = sensor;
}


/**
 * Gets object shape sensor.
 *
 * @return SoNodeSensor * Pointer to the object shape sensor.
 */
SoNodeSensor * SoShadowCaster::getShapeSensor()
{
  return this->shapeSensor;
}


/**
 * Create indexed Face Set model of the object in it's local coords.
 */
void SoShadowCaster::createFaceModel()
{
  objectModel->createFaceModel(this->objectRoot);
#if SHOWFACEMODEL
  this->showFaceModel();
#endif
}


/**
 * Adds object model to the scene graph.
 *
 * For debug purposes only.
 */
void SoShadowCaster::showFaceModel()
{
/*
  // delete old one
  if ( modelParent != NULL) 
    this->manager->getSceneRoot()->removeChild(modelParent);
*/
  SoTranslation *t1 = new SoTranslation;
  t1->translation.setValue(20.f, 0.f, 0.0f);
  SoSeparator * modelParent = new SoSeparator;
  modelParent->addChild(t1);

  // Rotate object
  SoRotationXYZ *rX = new SoRotationXYZ;
  rX->axis = SoRotationXYZ::X;
  rX->angle.setValue(float(3.1415926/2));
  modelParent->addChild(rX);

  modelParent->addChild(objectModel->getFaceModelNode());
  this->manager->getSceneRoot()->addChild(modelParent); // adds the model to scene
/*
// to write object model
  SoOutput output;
  output.openFile("object_model.iv");
  SoWriteAction * writer = new SoWriteAction(&output);
  writer->apply(modelParent);
  delete writer;
*/
}


/**
 * Finds and updates model's transformation matrix.
 * Stores previous one before.
 */
bool SoShadowCaster::getTransformationMatrix()
{
  // save old matrix
  this->trMatrixOld = trMatrix;

  SoSearchAction * sa = new SoSearchAction;
  sa->setNode(this->objectRoot);
  sa->setSearchingAll(FALSE);
  sa->apply(this->manager->getSceneRoot());

  SoPath * nodePath = sa->getPath();
  if ( nodePath == NULL ) {
    ONDEBUG( printf("MATRIX not FOUND\n") );
    return false;
  }

  const SbViewportRegion vpr;
  SoGetMatrixAction * getmatrixaction = new SoGetMatrixAction(vpr);
  getmatrixaction->apply(nodePath);
  trMatrix = getmatrixaction->getMatrix();
  ONDEBUG( printf("MATRIX FOUND\n") );

  return true;
}


/**
 * Set object move flag.
 */
SbMatrix * SoShadowCaster::getMatrix()
{
  return &this->trMatrix;
}


/**
 * Set object move flag.
 */
CObjectModel * SoShadowCaster::getObjectModel()
{
  return this->objectModel;
}


/**
 * Set object move flag.
 *
 * @param m Move flag.
 */
void SoShadowCaster::setMoved(bool m)
{
  this->hasMoved = m;
}


/**
 * Return object move flag.
 *
 * @return True if object has moved.
 */
bool SoShadowCaster::moved()
{
  return this->hasMoved;
}


/**
 * Re-computes all shadow volumes of this object and all lights.
 */
void SoShadowCaster::createShadowVolumes()
{
  ONDEBUG( printf("--SoShadowCaster::createShadowVolumes()\n") );

  // object has moved, get new transformation m.
  this->getTransformationMatrix();
  this->manager->createCasterVolumes(this->nameIndex);
}
