/**
 * @file
 * Shadow scene light class implementation.
 *
 * @author Tom "Shamot" Burian
 */
/* ---------------------------------------------------------------------------
 * Authors:      Tom Burian (tburian _AT centrum.cz)
 * Thanks to:    PCJohn (peciva _AT fit.vutbr.cz)
 * Contributors:
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 * ---------------------------------------------------------------------------
 */

//----------------------------------------------------------------------------
// INCLUDEs
//----------------------------------------------------------------------------

#include "SoShadowLight.h"
#include "SoShadowVolume.h"

//----------------------------------------------------------------------------
// IMPLEMENTATION
//----------------------------------------------------------------------------

/**
 * Run-time check if an OpenGL extension is present at the renderer.
 * 
 * @param extName Extension name string.
 * @return True if extension is found, false otherwise.
 */
GLboolean SoShadowLight::CheckExtension( char * extName )
{
  /*
   ** Search for extName in the extensions string.  Use of strstr()
   ** is not sufficient because extension names can be prefixes of
   ** other extension names.  Could use strtok() but the constant
   ** string returned by glGetString can be in read-only memory.
   */
  char *p = (char *) glGetString(GL_EXTENSIONS);
  char *end = NULL;
  int extNameLen;

  extNameLen = (int)strlen(extName);
  end = p + strlen(p);

  while (p < end) {
      int n = (int)strcspn(p, " ");
      if ((extNameLen == n) && (strncmp(extName, p, n) == 0)) {
          return GL_TRUE;
      }
      p += (n + 1);
  }
  return GL_FALSE;
}


GLenum SoShadowLight::incValue = GL_INCR;
GLenum SoShadowLight::decValue = GL_DECR;
 
float SoShadowLight::envIntensity = 0.0f;


/**
 * Checks if GL_EXT_stencil_wrap is present in run-time.
 */
void SoShadowLight::checkStencilExt()
{
/*
  int bits, max;   
  glGetIntegerv(GL_STENCIL_BITS, &bits);
  max = (1 << bits) - 1;
  printf("max stencil vals: %d\n",max);
*/

  // const GLubyte * extStr = glGetString(GL_EXTENSIONS);
  // printf("GL ext: %s\n",extStr);

/*
  GLenum incValue = GL_INCR;
  GLenum decValue = GL_DECR;
*/
  if ( CheckExtension("GL_EXT_stencil_wrap") )
  {
    incValue = GL_INCR_WRAP_EXT;
    decValue = GL_DECR_WRAP_EXT;
  }
  else
  {
    incValue = GL_INCR;
    decValue = GL_DECR;  
  }

}

/**
 * Constructor.
 *
 * @param sm Pointer connection to the Shadow Manager.
 * @param l Pointer connection to the light node.
 */
SoShadowLight::SoShadowLight(CShadowManager * sm, SoLight * l)
    : manager(sm),
      light(l)
{  
  nameIndex = -1;
  status = TRUE;

  pstatus = &this->status;

  lightRenderRoot = new SoSeparator;
  lightRenderRoot->setName("lightRenderRoot");
  lightRenderRoot->ref();  

  // switchNode = new SoSwitch(10); // problem with z-fail

  volumesRoot = new SoSeparator;
  volumesRoot->setName("volumesRoot");
  volumesRoot->ref();

  init();  
}


/**
 * Destructor.
 */
SoShadowLight::~SoShadowLight()
{
  // sensor must be detached 
  this->sensor->detach();
  this->removeAllChildren();

  lightRenderRoot->unref();

  volumesRoot->unref();
}


/**
 * Shadow light object initialization.
 *
 * Creates shadow rendering subgraph.
 */
void SoShadowLight::init()
{

  // problem with z-fail
/*
  switchNode->whichChild.setValue(SO_SWITCH_ALL);
     

  struct TLightSwitch * switchPair = new struct TLightSwitch;
  switchPair->switchNode = this->switchNode;
  switchPair->status = this->pstatus;

  SoCallback * lightStatusCallback = new SoCallback;
  lightStatusCallback->setCallback(lightStatusCB, switchPair );

  this->addChild(lightStatusCallback);
  this->switchNode->addChild(lightRenderRoot);
  this->addChild(switchNode);
*/
  // instead of to SoSwitch add lightRoot direct to this
  this->addChild(lightRenderRoot);


/*  
//
  // testy hodnot pro polygon offset
  pOffset->factor.setValue(0.1f); // 0.01? martyn
  pOffset->factor.setValue(0.1f); // 0.1 - TB vyzkouseno
  
  pOffset->factor.setValue(1.f); // test - asi ano, ale pro simple_ship se delaji hrany :(

  pOffset->units.setValue(50.0f); // 
  pOffset->units.setValue(50.0f); // TB - vypada OK

  // stadion
  pOffset->factor.setValue(0.01f); // 0.01? martyn
  pOffset->units.setValue(50.0f); // 1
//
*/
  // Polygon offset (need for zfail) must be set exactly like this
  SoPolygonOffset * pOffset = new SoPolygonOffset;

  pOffset->factor.setValue(0.1f); // default
  pOffset->units.setValue(50.0f); // default

  pOffset->styles.setValue(SoPolygonOffset::FILLED);

  // rendering callbacks
  SoCallback * incrementStencilCallback = new SoCallback;
  incrementStencilCallback->setCallback(incrementStencilCB, this->manager->getAlgorithm() ); // this->manager->getAlgorithm2()   //this->settings

  SoCallback * decrementStencilCallback = new SoCallback;
  decrementStencilCallback->setCallback(decrementStencilCB, this->manager->getAlgorithm() ); //this->settings

  SoCallback * shadowCallback = new SoCallback;

  struct TLightInfo * linfo = new struct TLightInfo;
  linfo->light = this->light;
  linfo->status = this->pstatus;
  linfo->globEnv = this->manager->getEnvironmentNode();
  shadowCallback->setCallback(litStencilCB, linfo);

  SoCallback * finishStencilCallback = new SoCallback;
  finishStencilCallback->setCallback(finishStencilCB, linfo);
  // end of rendering callbacks

  // Create a super root with callbacks and add scene, shadows and 
  // callbacks to switch stencil buffer settings for rendering shadows.

  //
  // create shadows scene graph
  //
  if ( this->manager->readSettings()->dropShadows ) // create shadows
  {           
    this->lightRenderRoot->addChild(incrementStencilCallback);        
    this->lightRenderRoot->addChild(pOffset);
    this->lightRenderRoot->addChild(this->volumesRoot);
    this->lightRenderRoot->addChild(decrementStencilCallback);
    this->lightRenderRoot->addChild(this->volumesRoot);
    this->lightRenderRoot->addChild(shadowCallback);
    this->lightRenderRoot->addChild(this->manager->getSceneRoot());
    this->lightRenderRoot->addChild(finishStencilCallback);    
  } 
  else // shadow volumes only, no shadows
  {    
    this->lightRenderRoot->addChild(this->manager->getSceneRoot());
    this->lightRenderRoot->addChild(pOffset);
    this->lightRenderRoot->addChild(this->volumesRoot);
  } 
}


/**
 * Returns pointer to the Shadow Manager.
 *
 * @return Pointer to the Shadow Manager.
 */
CShadowManager * SoShadowLight::getManager()
{
  return this->manager;
}


/**
 * Returns pointer to the light node.
 *
 * @return Pointer to the light node.
 */
SoLight * SoShadowLight::getLight()
{
  return this->light;
}


/**
 * Sets caster name index and the name of the node.
 *
 * @param lightIndex Auto increment index name.
 */
void SoShadowLight::setLightName(int lightIndex)
{
  SbString index(lightIndex);
  SbString name = "Light_";
  name += index;
  ONDEBUG( printf("-- light name: %s\n",name.getString()) );
  this->nameIndex = lightIndex;
  this->setName(name);
}


/**
 * Gets caster name index.
 *
 * @return int Index (name) of the object.
 */
int SoShadowLight::getLightName()
{
  return this->nameIndex;
}


/**
 * Sets light sensor.
 *
 * @param sensor Pointer to the PathSensor to the light node.
 */
void SoShadowLight::setSensor(SoNodeSensor * sensor)
{
  this->sensor = sensor;
}


/**
 * Gets light sensor.
 *
 * @return SoPathSensor * Pointer to the light sensor.
 */
SoNodeSensor * SoShadowLight::getSensor()
{
  return this->sensor;
}



/**
 * Add a new shadow volume to the light.
 *
 * @param volume A separator node of the shadow volume.
 */
void SoShadowLight::addVolume(SoSeparator * volume)
{
  ONDEBUG( printf("--SoShadowLight::addVolume()\n") );
  this->volumesRoot->addChild(volume);
}


/**
 * Deletes all shadow volumes under this light.
 */
void SoShadowLight::deleteVolumes()
{
  ONDEBUG( printf("--SoShadowLight::deleteVolumes()\n") );
  this->volumesRoot->removeAllChildren();
}


/**
 * Deletes shadow volumes i under this light.
 *
 * @param casterName Caster index name.
 */
void SoShadowLight::deleteVolume(int casterName)
{
  ONDEBUG( printf("--SoShadowLight::deleteVolume()\n") );
  int cnt = this->volumesRoot->getNumChildren();
  if ( cnt > 0 )
  {
    for ( int j = 0; j < cnt; j++ )
    {
      SoShadowVolume * volume = static_cast<SoShadowVolume *> (this->volumesRoot->getChild(j));
      if (volume->getCasterName() == casterName)
      {
        this->volumesRoot->removeChild(j);
      }
    }
  }

}


/**
 * Get number of shadow volumes under this light.
 *
 * @return Number of shadow volumes.
 */
int SoShadowLight::getNumVolumes()
{
  return this->volumesRoot->getNumChildren();
}


/**
 * Get shadow volume i.
 *
 * @param i Index of volume node to return.
 *
 * @return Pointer to the shadow volume node.
 */
SoSeparator * SoShadowLight::getVolume(int i)
{
  SoSeparator * volume = static_cast<SoSeparator *> (this->volumesRoot->getChild(i));
  return volume;
}


/**
 * Creates all shadow volumes of all objects for this light.
 */
void SoShadowLight::createShadowVolumes()
{
  ONDEBUG( printf("--SoShadowLight::createShadowVolumes()\n") );


  int cnt = this->volumesRoot->getNumChildren();
  ONDEBUG( printf("--SoShadowLight:: vol. root cnt: %d\n",cnt) );
  for (int i = 0; i < cnt; i++) // create SV for one light
  {
    SoShadowVolume * volume = static_cast<SoShadowVolume *>(this->volumesRoot->getChild(i));
    volume->create(); // this->nameIndex -1
  }

}

/**
 * Creates all shadow volumes of all objects for this light.
 *
 * @param caster Caster index name.
 */
void SoShadowLight::createShadowVolumes(int caster)
{
  ONDEBUG( printf("--SoShadowLight::createShadowVolumes()\n") );


  int cnt = this->volumesRoot->getNumChildren();
  ONDEBUG( printf("--SoShadowLight:: vol. root cnt: %d\n",cnt) );
  for (int i = 0; i < cnt; i++) // create SV for one light
  {
    SoShadowVolume * volume = static_cast<SoShadowVolume *>(this->getChild(i));
    if (volume->getCasterName() == caster )
    {
      volume->create();
      break;
    }
  }

}


//////////////////////////////////////////////////////////////////////////////
// Shadow rendering callbacks
//////////////////////////////////////////////////////////////////////////////

/*
 * According to the light status enables/disables rendering traversal 
 * of shadow volumes under the light.
 *
 * Sets stencil test according to the chosen algorithm.
 *
 * @param userdata Pointer to the pair of saved light status and the switch node.
 * @param action Pointer to the callback action.
 */
/*
void SoShadowLight::lightStatusCB(void *userdata, SoAction *action)
{ 
  if (action->isOfType(SoGLRenderAction::getClassTypeId()))
  {       
    struct TLightSwitch * switchPair = (TLightSwitch *) userdata;
    ONDEBUG(printf("lightStatusCB light is %d\n", *switchPair->status ));
    if ( *switchPair->status == TRUE )
    {
      switchPair->switchNode->whichChild.enableNotify(FALSE);
      switchPair->switchNode->whichChild.setValue(SO_SWITCH_ALL);
      switchPair->switchNode->whichChild.enableNotify(TRUE);
      // check stencil extension
      checkStencilExt();
    }
    else
    {
      switchPair->switchNode->whichChild.enableNotify(FALSE);
      switchPair->switchNode->whichChild.setValue(SO_SWITCH_NONE);
      switchPair->switchNode->whichChild.enableNotify(TRUE);
    }
  }
}
*/


/**
 * Rendering ++ CB.
 *
 * Sets stencil test according to the chosen algorithm.
 *
 * @param userdata Pointer to TShadowAlgorithm.
 * @param action Pointer to the callback action.
 */
void SoShadowLight::incrementStencilCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("incrementStencilCB\n"));    

    // check stencil extension
    checkStencilExt();


    glPopAttrib();
    
    glPushAttrib(GL_ALL_ATTRIB_BITS);

    glClear( GL_STENCIL_BUFFER_BIT );    
    
    glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
    glShadeModel(GL_FLAT);
    glDepthMask(GL_FALSE);
    glDepthFunc(GL_LESS);
    glEnable(GL_STENCIL_TEST);

    enum TShadowAlgorithm * algorithm = (TShadowAlgorithm *) userdata;
    if ( *algorithm == zfailAlg ) {
      ONDEBUG(printf("incrementStencilCB: zfail\n"));
      // z-fail
      //Increment stencil buffer for back face depth fail
      glStencilFunc(GL_ALWAYS, 0, ~0);
      // glStencilOp(GL_KEEP, GL_INCR, GL_KEEP);
      glStencilOp(GL_KEEP, incValue, GL_KEEP);
      glCullFace(GL_FRONT);
    } else {
      ONDEBUG(printf("incrementStencilCB: zpass\n"));
      // z-pass
      //Increment stencil buffer for front face depth pass
      glStencilFunc(GL_ALWAYS, 0, ~0);
      // glStencilOp(GL_KEEP, GL_KEEP, GL_INCR);
      glStencilOp(GL_KEEP, GL_KEEP, incValue);
      glCullFace(GL_BACK);
    }
  }
} 


/**
 * Rendering -- CB.
 *
 * Sets stencil test according to the chosen algorithm.
 *
 * @param userdata Pointer to TShadowAlgorithm.
 * @param action Pointer to the callback action.
 */
void SoShadowLight::decrementStencilCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("decrementStencilCB\n"));

    enum TShadowAlgorithm * algorithm = (TShadowAlgorithm *) userdata;
    if ( *algorithm == zfailAlg ) {
      ONDEBUG(printf("decrementStencilCB: zfail\n"));
      // z-fail
      // Decrement stencil buffer for back face depth pass
      // glStencilOp(GL_KEEP, GL_DECR, GL_KEEP);
      glStencilOp(GL_KEEP, decValue, GL_KEEP);
      glCullFace(GL_BACK);
    } else {
      ONDEBUG(printf("decrementStencilCB: zpass\n"));
      // z-pass
      // Decrement stencil buffer for front face depth pass
      // glStencilOp(GL_KEEP, GL_KEEP, GL_DECR);
      glStencilOp(GL_KEEP, GL_KEEP, decValue);
      glCullFace(GL_FRONT);
    }
  }
} 


/**
 * Rendering lit CB.
 * 
 * Turns off ambient light and draws the user scene beside shadowed
 * regions with one active light turned on. Shadow volumes of this light are 
 * tested with stencil buffer.
 * 
 * @param userdata Pointer to the pair of active light, all others lights,
 *        and ambient intensity field.
 * @param action Pointer to the callback action.
 */
void SoShadowLight::litStencilCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {

    ONDEBUG(printf("litStencilCB\n"));

    glPopAttrib();

    // Draw lit where stencil == 0

    glPushAttrib(GL_ENABLE_BIT);

    struct TLightInfo * linfo = (struct TLightInfo * ) userdata;

    // disable ambient light (additive blending)
    // (save ambient int. value)
    if ( linfo->globEnv != NULL )
    {
      linfo->globEnv->ambientIntensity.enableNotify(FALSE);
      envIntensity = linfo->globEnv->ambientIntensity.getValue();
      linfo->globEnv->ambientIntensity.setValue(0.0f);
      linfo->globEnv->ambientIntensity.enableNotify(TRUE);
    }

    // turn on the light (set the light to its saved value on/off)
    linfo->light->enableNotify(FALSE);
    linfo->light->on = *linfo->status;    
    linfo->light->enableNotify(TRUE);

    glEnable(GL_LIGHTING);

    glBlendFunc(GL_ONE, GL_ONE);
    glEnable(GL_BLEND);    
   
    // set stencil test
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
    glDepthFunc(GL_EQUAL);
    glStencilFunc(GL_EQUAL, 0, ~0);
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
    glEnable(GL_STENCIL_TEST);

  }
} 


/**
 * Rendering finish CB.
 * 
 * Turns on ambient light and turns off the active light. 
 * 
 * @param userdata Pointer to the pair of active light, all others lights,
 *        and ambient intensity field.
 * @param action Pointer to the callback action.
 */
void SoShadowLight::finishStencilCB(void *userdata, SoAction *action)
{  
  if (action->isOfType(SoGLRenderAction::getClassTypeId())) {
    ONDEBUG(printf("finishStencilCB\n"));

    glPopAttrib();    

    glDisable(GL_LIGHTING);

    glDisable(GL_STENCIL_TEST);

    glDisable(GL_BLEND);

    glPushAttrib(GL_ENABLE_BIT);
        
    struct TLightInfo * linfo = (struct TLightInfo * ) userdata;

    // enable ambient light 
    if ( linfo->globEnv != NULL )
    {
      linfo->globEnv->ambientIntensity.enableNotify(FALSE);
      linfo->globEnv->ambientIntensity.setValue(envIntensity);
      linfo->globEnv->ambientIntensity.enableNotify(TRUE);
    }

    // turn off the light
    linfo->light->enableNotify(FALSE);
    linfo->light->on = FALSE;
    linfo->light->enableNotify(TRUE);    
  }
} 

