/**
 * @file
 * Shadow scene light class definition.
 *
 * @author Tom "Shamot" Burian
 */
/* ---------------------------------------------------------------------------
 * Authors:      Tom Burian (tburian _AT centrum.cz)
 * Thanks to:    PCJohn (peciva _AT fit.vutbr.cz)
 * Contributors:
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 * ---------------------------------------------------------------------------
 */

#ifndef SO_SHADOW_LIGHT__H
#define SO_SHADOW_LIGHT__H


#ifdef _WIN32
  #include <windows.h> // needed by gl.h
#endif

// Must be linked with glu32.lib opengl32.lib
#include <GL/gl.h>

/**
\note 
OpenGL needed extension: EXT_stencil_wrap.
This extension solves problem when more than 2^n+1 values are rendered
into stencil buffer. It wraps values around.
See http://developer.3dlabs.com/EXTmanpages/stencil_wrap.txt for details.

*/

/** Address of the enum value, if not in gl.h. */
#define GL_INCR_WRAP_EXT             0x8507 
/** Address of the enum value, if not in gl.h. */
#define GL_DECR_WRAP_EXT             0x8508 


//----------------------------------------------------------------------------
// INCLUDEs
//----------------------------------------------------------------------------

#include "ShadowManager.h"
//
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoSwitch.h>
#include <Inventor/nodes/SoNormalBinding.h>
#include <Inventor/nodes/SoShapeHints.h>
#include <Inventor/nodes/SoIndexedFaceSet.h>
#include <Inventor/nodes/SoTranslation.h>
//
#include <Inventor/actions/SoCallbackAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/actions/SoGetMatrixAction.h>
//
#include <Inventor/elements/SoShapeHintsElement.h>
//
#include <Inventor/misc/SoChildList.h>
//
#include <Inventor/SoPrimitiveVertex.h>
#include <Inventor/SbLinear.h>

//----------------------------------------------------------------------------
// IMPLEMENTATION
//----------------------------------------------------------------------------

/**
 * Class defining a light in a shadow scene. 
 *
 * An instance of this class is created for every light added into 
 * the Shadow Manager and id is appended to shadowLightsRoot.
 * SoShadowLight contains a part of a final shadow rendering scene graph
 * that renders shadows created by this light.
 * If the light is turned off (i.e. casts no shadows) the scene
 * graph below is not traversed.
 */
class SoShadowLight : public SoSeparator
{
  private:
    /** Pointer to shadow manager. */
    CShadowManager * manager;
    /** Light. */
    SoLight * light;
    /** Light render root. */
    SoSeparator * lightRenderRoot;
    /** Shadow volumes root node (vol. created by this light). */
    SoSeparator * volumesRoot;
    /** Switch node above the volumes root. */
    //SoSwitch * switchNode; // problem with z-fail
    /** Sensor of light movement or change. */
    SoNodeSensor * sensor;
    /** Object name index. */
    int nameIndex;
    /** Ambient light intensity saved value. */
    static float envIntensity;    
    /** OpenGL enum for incrementing stencil buf. Depends on if EXT_stencil_wrap extension is present. */
    static GLenum incValue;
    /** OpenGL enum for decrementing stencil buf. Depends on if EXT_stencil_wrap extension is present. */
    static GLenum decValue;

    static void checkStencilExt();
    static GLboolean CheckExtension( char *extName );

    void init();

// shadow rendering callbacks
    // static void lightStatusCB(void * userdata, SoAction * action); // problem with z-fail
    static void incrementStencilCB(void * userdata, SoAction * action);
    static void decrementStencilCB(void * userdata, SoAction * action);
    static void litStencilCB(void * userdata, SoAction * action);
    static void finishStencilCB(void * userdata, SoAction * action);    

  public:
    /** Saved light status. */
    SbBool status;
    /** Pointer to saved light status. */
    SbBool * pstatus;

    SoShadowLight(CShadowManager * sm, SoLight * l);
    ~SoShadowLight();

    void addVolume(SoSeparator * volume);
    SoSeparator * getVolume(int i);
    int getNumVolumes();
    void deleteVolumes();
    void deleteVolume(int casterName);    

    CShadowManager * getManager();

    SoLight * getLight();
    void setLightName(int lightIndex);
    int getLightName();

    void setSensor(SoNodeSensor * sensor);
    SoNodeSensor * getSensor();
    void createShadowVolumes();
    void createShadowVolumes(int caster);

};


#endif
