/**
 * @file
 * Shadow volumes class implementation.
 *
 * @author Tom "Shamot" Burian
 */
/* ---------------------------------------------------------------------------
 * Authors:      Tom Burian (tburian _AT centrum.cz)            
 * Thanks to:    PCJohn (peciva _AT fit.vutbr.cz)
 * Contributors: 
 *
 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 * This source code is offered for use in the public domain.
 * You may use, modify or distribute it freely.
 *
 * This source code is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY
 * DISCLAIMED.  This includes but is not limited to warranties of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * If you find the source code useful, authors will kindly welcome
 * if you give them credit and keep their names with their source code,
 * but do not feel to be forced to do so.
 * ---------------------------------------------------------------------------
 */

//----------------------------------------------------------------------------
// INCLUDEs
//----------------------------------------------------------------------------

#include "SoShadowVolume.h"

//----------------------------------------------------------------------------
// IMPLEMENTATION
//----------------------------------------------------------------------------

/**
 * Constructor.
 *
 * @param l      Pointer to light that casts this volume.
 * @param cn     Pointer to shadow caster that casts this volume.
 */
SoShadowVolume::SoShadowVolume(SoShadowLight *l, int cn)
    : shadowlight(l),
      casterName(cn)
{

  transform = new SoTransform();
  nScTr = 0;
  nSVTr = 0;  
  aLight = shadowlight->getLight();

  objectModel = NULL;
  trMatrix = NULL;
  caster = NULL;

  if ( casterName >= 0 ) // find caster node
  {
    // caster node name
    SbString index(casterName);
    SbString name = "Caster_";
    name += index;    

    // get and set actual information from caster
    caster = (SoShadowCaster *)SoNode::getByName(name);
    ONDEBUG( printf("SV::constructor caster name (%d): %d\n", casterName, caster->getCasterName()) );
  }

  initNode = NULL;
  volumeNode = NULL;

  this->init();
}


/**
 * Destructor.
 */
SoShadowVolume::~SoShadowVolume()
{
  this->removeAllChildren();
}


/**
 * Initializator.
 */
void SoShadowVolume::init()
{ 
  this->initNode = new SoGroup;
  this->addChild(initNode);
  this->volumeNode = new SoGroup;      
  this->addChild(volumeNode);

  // transformation used for directional light
  // has no sense in other cases 
  transform->translation.setValue(0.0f, 0.0f, 0.0f);
  transform->rotation.setValue(0.0f, 0.0f, 0.0f, 1.0f);
  transform->scaleFactor.setValue(1.0f, 1.0f, 1.0f);
  transform->scaleOrientation.setValue(0.0f, 0.0f, 0.0f, 1.0f);
  transform->center.setValue(0.0f, 0.0f, 0.0f);
  this->initNode->addChild(transform);

  // shadow volumes material
  SoMaterial *material = new SoMaterial;
  material->ambientColor.setValue(svColor);
  material->diffuseColor.setValue(svColor);
//  if ( ! settings.dropShadows )    material->transparency.setValue(0.1f);
  this->initNode->addChild(material);

  // shadow shape hints
  SoShapeHints * hints = new SoShapeHints;
  hints->shapeType = SoShapeHintsElement::SOLID;
  hints->vertexOrdering = SoShapeHintsElement::COUNTERCLOCKWISE;
  this->initNode->addChild(hints);

  // normal binding
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_FACE_INDEXED);
  this->initNode->addChild(nbind);
}


/**
 * Return the asociated caster name index.
 *
 * @return Caster name index.
 */
int SoShadowVolume::getCasterName()
{
  return this->casterName;
}


/**
 * Creates object's shadow volume for one light.
 */
void SoShadowVolume::create()
{
  ONDEBUG( printf("+++SoShadowVolume::create()\n") );

  if ( aLight != NULL ) 
  {         
    // check light status
    SbBool lightON = TRUE;  
    aLight->on.enableNotify(FALSE);  
    lightON = this->aLight->on.getValue();
    aLight->on.enableNotify(TRUE);

    ONDEBUG(
      if ( lightON )
       printf("+++SoShadowVolume::light is ON\n");
      else
        printf("+++SoShadowVolume::light is OFF\n");
    );

    // optimalisation for directional light
    // no need to recompute shadow volumes when object has moved
    // just compute translation vector and add it to the SV's transformation node
    ONDEBUG( printf("caster name: %d\n",casterName) );
    SbVec3f translation(0.0f, 0.0f, 0.0f);
    bool dirOpt = FALSE; // can be directional light optimalisation used?

    if ( caster != NULL ) // find parent caster node, if object changes
    {
      trMatrix = caster->getMatrix();
      objectModel = caster->getObjectModel();

      if ( caster->moved() )
      {
        ONDEBUG( printf("CASTER MOVEMENT: \n") );
        if (   aLight->getTypeId() == SoDirectionalLight::getClassTypeId() 
            || aLight->getTypeId() == SoDirectionalLightManip::getClassTypeId()
           )
        {
          ONDEBUG( printf("dir. light optimalisation \n") );
          SbVec3f pointOld(1.0f, 1.0f, 1.0f);
          SbVec3f pointNew(1.0f, 1.0f, 1.0f);
          caster->trMatrixOld.multVecMatrix(pointOld, pointOld);
          caster->trMatrix.multVecMatrix(pointNew, pointNew);
          translation = pointNew - pointOld;
          dirOpt = TRUE;
        }      
        caster->setMoved(FALSE);
      }
    }
  

    if ( !dirOpt )
    {
      // delete old shadow volume's coordinates etc.; clear the volume node      
      ONDEBUG( 
        int x = this->volumeNode->getNumChildren();
        printf("volume had %d children\n",x);
      );

      // delete all shadow volumes under this node
      this->volumeNode->removeAllChildren();                     

      nScTr = 0;
      nSVTr = 0;  

      transform->translation.setValue(0.0f, 0.0f, 0.0f);
    }

    else
    {
      // add the translation
      transform->translation.setValue(transform->translation.getValue() + translation);
    }

    if ( lightON && !dirOpt ) // re-compute shadow volume only if the light is turned ON
                              // and it is not a directional light (no optimalisation)
    {
      //this->init();

      // determine algorithm
      this->shadowlight->getManager()->method2algorithm();
 
      int itn = 0; // index to normal indices array
      int j = 2; // actual triangle in the strip set, vertices: j-2, j-1 and j
      int itmat = 0;
      float vysl = 0.0f;

      SbVec3f normal;
      // Debug: Check if we are in the range
      int b = this->objectModel->modelIfs->normalIndex.getNum();

      // Create a shadow volume for each light-facing triangle
      // Goes over all triangles in the triangleStripSet
      while ( j < objectModel->modelIfs->coordIndex.getNum() ) {

        while ( objectModel->modelIfs->coordIndex[j] != -1 ) {

          if ( b < itn ) {
            ONDEBUG( fprintf(stderr,"Normal index out of range.\n\n") );
            exit(1);
          }

          // Get the normal vector
          int nindex = objectModel->modelIfs->normalIndex[itn];
          normal = objectModel->modelNormal->vector[nindex];            

          // Transformation of a vertex and normal vector to global coordinates
          SbVec3f x(objectModel->modelCoord3->point[objectModel->modelIfs->coordIndex[j-2]]);
          SbVec3f normEnd(x+1*normal);  // endpoint of the normal vector

          // Vertex
          trMatrix->multVecMatrix(x, x);
          // New normal vector
          trMatrix->multVecMatrix(normEnd, normEnd);
          normal = normEnd - x;

          if ( normal.length() > .0f ) {
            normal.normalize();
          } else {
            ONDEBUG(printf("Error: Length of the normal is 0.\n"));
          }

          // lights: 
          // according the light type compute light location and light direction vector
          SbVec3f lightLocation;
          SbVec3f lightDirection;
          enum TLightTypes lightType;
          char * typ = "";
      
          if (aLight->getTypeId() == SoDirectionalLight::getClassTypeId() )
          {
            lightType = TSoDirectionalLight;
            typ = "TSoDirectionalLight";
            SoDirectionalLight * light = (SoDirectionalLight *)aLight;
            lightDirection = light->direction.getValue();
          }
          else if (aLight->getTypeId() == SoDirectionalLightManip::getClassTypeId() )
          {
            lightType = TSoDirectionalLightManip;
            typ = "TSoDirectionalLightManip";
            SoDirectionalLightManip * light = (SoDirectionalLightManip *)aLight;
            lightDirection = light->direction.getValue();
          }
          else if (aLight->getTypeId() == SoPointLight::getClassTypeId() )
          {
            lightType = TSoPointLight;
            typ = "TSoPointLight";
            SoPointLight * light = (SoPointLight *)aLight;
            lightLocation = light->location.getValue();
          }
          else if (aLight->getTypeId() == SoPointLightManip::getClassTypeId() )
          {
            lightType = TSoPointLightManip;
            typ = "TSoPointLightManip";
            SoPointLightManip * light = (SoPointLightManip *)(aLight);
            lightLocation = light->location.getValue();
          }
          else if (aLight->getTypeId() == SoSpotLight::getClassTypeId() )
          {
            lightType = TSoSpotLight;
            typ = "TSoSpotLight";
            SoSpotLight * light = (SoSpotLight *)(aLight);
            lightLocation = light->location.getValue();
          }
          else if (aLight->getTypeId() == SoSpotLightManip::getClassTypeId() )
          {
            lightType = TSoSpotLightManip;
            typ = "TSoSpotLightManip";
            SoSpotLightManip * light = (SoSpotLightManip *)aLight;
            lightLocation = light->location.getValue();
          }
          else 
          {
            lightType = TUnknownLightType;
            typ = "TUnknownLightType";
          }

          // light direction vector. it's a vector from the vertex 
          // to light source (point and spot light) or light direction
          // (directional light)
          SbVec3f lv;
          if (lightType == TSoDirectionalLight)
          {
            lv = lightDirection;
          }
          else if (lightType == TSoDirectionalLightManip)
          {                
            lv = lightDirection;
          }
          else if (lightType == TSoPointLight)
          {
            lv = (lightLocation-x);
          }
          else if (lightType == TSoPointLightManip)
          {                
            lv = (lightLocation-x);
          }
          else if (lightType == TSoSpotLight)
          {                
            lv = (lightLocation-x);
          }
          else if (lightType == TSoSpotLightManip)
          {
            lv = (lightLocation-x);
          }          

          // Dot product of normal and light direction vector
          // to determine if the triangle is light facing or not
          vysl = normal.dot(lv);
          if ( vysl <= 0.0f )
          {
            // Not facing light
            int n = objectModel->modelIfs->materialIndex.getNum();
            objectModel->modelIfs->materialIndex.set1Value(n,0);
            objectModel->modelIfs->materialIndex.setNum(n+1);
          } 
          else 
          {
            // Is light-facing, part of the silhouette
            this->nScTr++;
            int n = objectModel->modelIfs->materialIndex.getNum();
            objectModel->modelIfs->materialIndex.set1Value(n,1);
            objectModel->modelIfs->materialIndex.setNum(n+1);

            // Create envelope of the shadow volume

            SoCoordinate3 *envCoords = new SoCoordinate3;
            SoIndexedTriangleStripSet * envStrip = new SoIndexedTriangleStripSet;
            // Note: normals are computed automatically by Coin according to the 
            // SoNormalBinding::PER_FACE_INDEXED from the coordinates

            envCoords->point.setNum(2*3+2); // 8 = 2 * #edges + 2 (closed)
            this->nSVTr += 6; // #tr of SV envelope

            SbVec3f *c = envCoords->point.startEditing();

            int k = 0;

            // compute verteces of the shadow volume
            for ( k=3; k >= 0; k-- )
            {
              // To global coordinates
              int t = j-k%3;
              SbVec3f xx(objectModel->modelCoord3->point[objectModel->modelIfs->coordIndex[t]]);
              trMatrix->multVecMatrix(xx, xx);

              SbVec3f l2pVec; // light dir. vector, depends on light type
              if (lightType != TSoDirectionalLight && lightType != TSoDirectionalLightManip)
              {
                l2pVec = xx - lightLocation;
              }
              else
              {
                l2pVec = lightDirection;
              }
         
              // point of the far cap, comp. by using light direction vector and a triangle vertex
              SbVec3f xxFar(farPointNew(l2pVec, xx, this->shadowlight->getManager()->readSettings()->fDepth));
              *(c++) = xx;
              *(c++) = xxFar;  // Far point
            }
            envCoords->point.finishEditing();

            // Add nodes to scene graph
            this->volumeNode->addChild(envCoords);
        
            // puvodni SoIndexedTriangleStripSet * envStrip = new SoIndexedTriangleStripSet;
            int32_t indeces[24] = {0,1,2,-1,2,1,3,-1, 2,3,4,-1,4,3,5,-1, 4,5,6,-1,6,5,7,-1};
            // for directional light indices must be exact contrariwise than for point & spot light
            // do not know why
            int32_t indecesDir[24] = {1,0,2,-1, 1,2,3,-1, 2,4,3,-1, 3,4,5,-1, 4,6,5,-1, 6,7,5,-1};
            if (lightType == TSoDirectionalLight || lightType == TSoDirectionalLightManip)
              envStrip->coordIndex.setValues(0,24,indecesDir);
            else
              envStrip->coordIndex.setValues(0,24,indeces);
            this->volumeNode->addChild(envStrip);

            // TODO: determine algorithm
            // this->getManager()->method2algorithm();

            // cap volumes only if zfail used
            if ( *this->shadowlight->getManager()->getAlgorithm() == zfailAlg )
            // if ( this->getManager()->getSettings()->method == zfail )
            {     
              // Near & far cap of the volume
              SoIndexedTriangleStripSet * ncapStrip = new SoIndexedTriangleStripSet;
              int32_t ncapIndeces[4] = {0,2,4,-1};
              // dir. light
              int32_t ncapIndecesDir[4] = {0,4,2,-1};
              if (lightType == TSoDirectionalLight || lightType == TSoDirectionalLightManip)
                ncapStrip->coordIndex.setValues(0,4,ncapIndecesDir);
              else
                ncapStrip->coordIndex.setValues(0,4,ncapIndeces);
              this->volumeNode->addChild(ncapStrip);

              SoIndexedTriangleStripSet * fcapStrip = new SoIndexedTriangleStripSet;
              int32_t fcapIndeces[4] = {1,5,3,-1}; // 1,3,5
              // dir. light
              int32_t fcapIndecesDir[4] = {1,3,5,-1};
              if (lightType == TSoDirectionalLight || lightType == TSoDirectionalLightManip)
                fcapStrip->coordIndex.setValues(0,4,fcapIndecesDir);
              else
                fcapStrip->coordIndex.setValues(0,4,fcapIndeces);
              this->volumeNode->addChild(fcapStrip);

              //
              this->nSVTr += 2; // #tr of SV near and far cap
            }


          }
          // End of a strip, move iterators
          itn++;
          j++;

        } // while

        j += 3;  // end of strip (-1) => move to next one

      } // while coords
      ONDEBUG( printf("Shadow volumes of the object created. They consists of %d shadow casting tr.\n", this->nScTr) );
    } // end if light is ON

  } // aLight not NULL
  else
  {
    printf("light is NULL\n");
  }
}


/**
 * Writes generated shadow node scene graph into a file.
 *
 * @param filename Output filename.
 */
void SoShadowVolume::writeShadowVolume(const char *filename)
{
  SoOutput output;

  if (this->getChildren()->getLength() != 0) {
    output.openFile(filename);
    SoWriteAction * writer = new SoWriteAction(&output);
    writer->apply(this);
    delete writer;
    ONDEBUG( fprintf(stdout,"Shadows' scene graph was written in '%s'.\n\n", filename) );
  } else {
    ONDEBUG( fprintf(stderr,"Error: Could not write shadows' scene graph.\nShadow volumes have not been generated yet.\n\n") );
  }
}


/**
 * Computes shadow volume far plane point (far cap).
 *
 * Melo by se dodelat tak, ze bude az na zadni orez. rovine.
 *
 * @param lightPos Light position vector.
 * @param filename Vertex of a light facing triangle.
 */
/*
SbVec3f SoShadowVolume::farPoint(SbVec3f lightPos, SbVec3f point)
{
  SbVec3f farPoint(point);
  SbVec3f l2pVec(point-lightPos);
  l2pVec.normalize();

  float farDepth2 = 150.f;

  farPoint = point + (farDepth2 * l2pVec);

  return farPoint;
}
*/


/**
 * Computes shadow volume far plane point (far cap) - a new version.
 *
 * Melo by se dodelat tak, ze bude az na zadni orez. rovine.
 *
 * @param lightVec Light direction vector.
 * @param point Vertex of a light facing triangle.
 * @param farDepth Depth of a shadow volume.
 *
 * @return A point of a far cap of the shadow volume.
 */
SbVec3f SoShadowVolume::farPointNew(SbVec3f lightVec, SbVec3f point, float farDepth)
{
  SbVec3f farPoint(point);
  lightVec.normalize();
  farPoint = point + farDepth * lightVec;

  return farPoint;
}


/**
 * Computes the triangle normal vector from its verteces.
 *
 * Counterclockwise orientation.
 *
 * @param p1 point 1
 * @param p2 point 2
 * @param p3 point 3
 *
 * @return A normal vector.
 */
SbVec3f SoShadowVolume::faceNormal(const SbVec3f p1, const SbVec3f p2, const SbVec3f p3)
{
  SbVec3f vector1, vector2;

  vector1 = p2 - p1;
  vector2 = p3 - p2;

  if ( vector1.length() <= .0f ) {
    ONDEBUG(printf("Error: Length of a vector1 is 0. Correcting...\n"));
  }
  if ( vector2.length() <= .0f ) {
    ONDEBUG(printf("Error: Length of a vector2 is 0. Correcting...\n"));
  }

  SbVec3f normal = vector1.cross(vector2);
  if ( normal.length() > .0f ) {
    normal.normalize();
  } else {
    ONDEBUG(printf("Error: FaceNormal: Length of the normal is 0.\n"));
  }

  return normal;
}
